#!C:\bld\hendrics_1543255133425\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'hendrics==4.0rc1.post2','console_scripts','HENlags'
__requires__ = 'hendrics==4.0rc1.post2'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('hendrics==4.0rc1.post2', 'console_scripts', 'HENlags')()
    )
